clustering_threshold = 10
hashing_depth = 1
clusterize_using_dcup = False
report_unifiers = False
print_time = True
force = False  
use_diff = True 
clusterize_using_hash = False
links_for_eclipse = False
